<?php

namespace NOUT;

use NOUT\Entity\ConfirmResponse;
use NOUT\Entity\UsernameToken\Base64UsernameToken;
use NOUT\Entity\UsernameToken\CipherUsernameToken;
use \NOUT\Entity\UsernameToken\PartialOASISUsernameToken;
use NOUT\Entity\ActionLog;
use NOUT\Entity\InfoLog;
use NOUT\Entity\RequestLog;
use NOUT\Entity\UsernameToken\OASISUsernameToken;
use NOUT\Entity\UsernameToken\UsernameTokenInterface;
use NOUT\Soap\ClassMap;
use NOUT\Soap\EnumType\ReturnType;
use NOUT\Soap\PackageBase\SoapClientInterface;
use NOUT\Soap\ServiceType\SIMAXService;
use NOUT\Soap\StructType\Action;
use NOUT\Soap\StructType\Create;
use NOUT\Soap\StructType\Delete;
use NOUT\Soap\StructType\Disconnect;
use NOUT\Soap\StructType\Element;
use NOUT\Soap\StructType\EmptyType;
use NOUT\Soap\StructType\Execute;
use NOUT\Soap\StructType\GetTokenSession;
use NOUT\Soap\StructType\OptionDialogue;
use NOUT\Soap\StructType\Update;
use NOUT\Soap\StructType\Validate;
use NOUT\Soap\StructType\XmlType;

abstract class NOUTClient
{
    protected const APIUUID = '41206c18-e50d-4244-83ca-eddf8d10295b';

    const OPT_PORT = 'port';
    const OPT_HOST = 'host';
    const OPT_URL = 'url';
    const OPT_USERNAME = 'username';
    const OPT_PASSWORD = 'password';
    const OPT_MODEAUTH = 'mode_auth';
    const OPT_TOKEN = 'token';
    const OPT_SECRET = 'secret';
    const OPT_DISPLAY_VALUE = 'display_value';
    const OPT_RETURN_XSD = 'return_xsd';
    const OPT_VERSION_MIN = 'version_min';
    const OPT_VERSION_PREF = 'version_pref';
    const OPT_WITH_STATE_CONTROL = 'with_state_control';
    const OPT_APIUUID = 'apiuuid';
    const OPT_WSDL_FILE = 'wsdl_file';


    const ERROR_SIMAXONLINESECURITE = 9997;
    const ERROR_NEEDSOAPEXT         = 9998;
    const ERROR_NEEDCURLEXT         = 9996;
    const ERROR_NEEDAUTHINFORMATION = 9995;
    const ERROR_NEEDENTRYPOINT = 9994;

    const ERROR_BADCREDENTIALS = 10;
    const ERROR_TRAITERETOUR = 99;
    const ERROR_MESSAGEBOXNOTOK = 101;
    const ERROR_CONFIRMMESSAGEBOX = 200;
    const ERROR_OUVERTUREFICHE = 201;
    const ERROR_ONSAVE = 202;
    const ERROR_OUVERTUREPARAMETRE = 203;
    const ERROR_BADWORKFLOW = 204;


    /** @var array|null  */
    protected array $aOptions;

    /** @var UsernameTokenInterface|null */
    protected ?UsernameTokenInterface $clUsernameToken = null;

    /** @var OptionDialogue  */
    protected OptionDialogue $oOptionDialogue;

    /** @var null|string  */
    protected ?string $sTokenSession = null;

    /** @var null|DebugTrace  */
    protected ?DebugTrace $clDebugTrace = null;

    /** @var SIMAXService|null */
    protected ?SIMAXService $clSoapService = null;

    /** @var ActionLog|null  */
    protected ?ActionLog $clActionLog = null;

    /** @var null|InfoLog  */
    protected ?InfoLog $clInfoLog = null;

    /** @var string */
    protected string $sRequestID;

    /**
     * @param array|null $options
     */
    public function __construct(array $options = null)
    {
        $this->sRequestID = uniqid('request-');
        $this->aOptions   = array(
            self::OPT_MODEAUTH => 'oasis',
            self::OPT_URL => null,
            self::OPT_HOST => null,
            self::OPT_PORT => 8052,
            self::OPT_USERNAME => '',
            self::OPT_PASSWORD => '',
            self::OPT_TOKEN => null,
            self::OPT_SECRET => file_get_contents(__DIR__. '/defaultSecret.txt'),
            self::OPT_DISPLAY_VALUE => 0,
            self::OPT_RETURN_XSD => 1,
            self::OPT_VERSION_MIN => (PHP_MAJOR_VERSION < 8 || PHP_MAJOR_VERSION == 8 && PHP_MINOR_VERSION < 2) ? 1 : 14,
            self::OPT_VERSION_PREF => (PHP_MAJOR_VERSION < 8 || PHP_MAJOR_VERSION == 8 && PHP_MINOR_VERSION < 2) ? 1 : 14,
            self::OPT_WITH_STATE_CONTROL => 0,
            self::OPT_APIUUID => self::APIUUID,
            self::OPT_WSDL_FILE => null,
        );

        if ($options) {
            $this->aOptions = $options + $this->aOptions;
            if (empty($this->aOptions[self::OPT_SECRET])) {
                $this->aOptions[self::OPT_SECRET] = file_get_contents(__DIR__. '/defaultSecret.txt');
            }
            if (empty($this->aOptions[self::OPT_APIUUID])){
                $this->aOptions[self::OPT_APIUUID] = self::APIUUID;
            }
        }
    }

    protected function _aInitBeforeConnexion() : array
    {
        //------
        //vérification des extensions php
        $bIsSoapExtOk = extension_loaded('soap');
        //$bIsCurlExtOk = extension_loaded('curl');
        if (!$bIsSoapExtOk/* || !$bIsCurlExtOk*/)
        {
            $message= 'need php extension soap';
            $this->_writeTrace($message);
            throw new \Exception($message, self::ERROR_NEEDSOAPEXT);
        }

        //
        if (empty($this->aOptions[self::OPT_URL]) && empty($this->aOptions[self::OPT_HOST]))
        {
            $message= 'need an entry point, either url or online address/port';
            $this->_writeTrace($message);
            throw new \Exception($message, self::ERROR_NEEDENTRYPOINT);
        }

        if (!empty($this->aOptions[self::OPT_URL])){
            $endpoint = $this->aOptions[self::OPT_URL];
        }
        else {
            if (intval($this->aOptions[self::OPT_PORT]) == 0){
                $message= 'need an entry point, either url or online address/port';
                $this->_writeTrace($message);
                throw new \Exception($message, self::ERROR_NEEDENTRYPOINT);
            }
            $endpoint = "http://{$this->aOptions[self::OPT_HOST]}:{$this->aOptions[self::OPT_PORT]}";
        }

        $wsdl_endpoint = rtrim($endpoint, '/').'/GetWSDL'.(!empty($this->aOptions[self::OPT_APIUUID]) ? "?!apiuuid={$this->aOptions[self::OPT_APIUUID]}" : '');

        //----------------
        //vérification des informations de connexions
        if (empty($this->aOptions[self::OPT_TOKEN])) {
            if (empty($this->aOptions[self::OPT_USERNAME])){
                throw new \Exception('no identification information', self::ERROR_NEEDAUTHINFORMATION);
            }

            $restClient = new Rest\Service($endpoint, null);
            $clUserExists = $restClient->clGetUserExists($this->aOptions[self::OPT_USERNAME]);

            switch ($this->aOptions[self::OPT_MODEAUTH]) {
                case 'base64':
                    $this->clUsernameToken = new Base64UsernameToken(
                        $this->aOptions[self::OPT_USERNAME],
                        $this->aOptions[self::OPT_PASSWORD],
                        $this->aOptions[self::OPT_SECRET]
                    );
                    break;
                case 'aes':
                case 'blowfish':
                    $this->clUsernameToken = new CipherUsernameToken(
                        $this->aOptions[self::OPT_MODEAUTH],
                        $this->aOptions[self::OPT_USERNAME],
                        $this->aOptions[self::OPT_PASSWORD],
                        $this->aOptions[self::OPT_SECRET]
                    );
                    break;
                default:
                    $this->clUsernameToken = new OASISUsernameToken(
                        $this->aOptions[self::OPT_USERNAME],
                        $this->aOptions[self::OPT_PASSWORD],
                        $clUserExists->clEncryptionType
                    );
                    break;
            }

            $this->clUsernameToken->Compute();
        }

        $this->oOptionDialogue = new OptionDialogue(0,
                                                    1,
                                                    1,
                                                    $this->aOptions[self::OPT_RETURN_XSD],
                                                    0,
                                                    0,
                                                    0,
                                                    $this->aOptions[self::OPT_DISPLAY_VALUE],
                                                    12,
                                                    $this->aOptions[self::OPT_WITH_STATE_CONTROL],
                                                    0,
                                                    $this->aOptions[self::OPT_VERSION_MIN],
                                                    $this->aOptions[self::OPT_VERSION_PREF]);

        return [$endpoint, $wsdl_endpoint];
    }

    /**
     * @return bool
     * @throws \Exception
     */
    public function bConnecte() : bool
    {
        if (!is_null($this->clSoapService)){
            return true; //on ne se connecte pas 2 fois
        }

        list($endpoint, $wsdlEnpoint) = $this->_aInitBeforeConnexion();

        if (strpos($wsdlEnpoint, 'https://') === 0){
            $wsdlcontent = @file_get_contents($wsdlEnpoint);
            if ($wsdlcontent === false)
            {
                $code = self::ERROR_SIMAXONLINESECURITE;
                $message="Error while reading the WSDL : Verify the uuid security.";
                $this->_writeTrace($message);
                throw new \Exception($message, $code);
            }

            $wsdlEnpoint = __DIR__.'/online.wsdl';
            file_put_contents($wsdlEnpoint, $wsdlcontent);
        }

        ini_set("default_socket_timeout", 36000);
        try{
            $options = array(
                SoapClientInterface::WSDL_SOAP_VERSION => SOAP_1_2,
                SoapClientInterface::WSDL_URL => $wsdlEnpoint,
                SIMAXService::WSDL_PROXY_URI => $endpoint,
                SoapClientInterface::WSDL_CLASSMAP => ClassMap::get(),
            );

            $oSOAPClient = new SIMAXService($options);
            $oSOAPClient->getSoapClient()->__setLocation($endpoint);
        }
        catch (\Exception $e)
        {
            $message = $e->getMessage();
            //SOAP-ERROR: Parsing WSDL: Couldn't load from 'http://127.0.0.1:8062/getwsdl' : failed to load external entity "http://127.0.0.1:8062/getwsdl"
            if ((strncasecmp($message, "SOAP-ERROR: Parsing WSDL:", strlen("SOAP-ERROR: Parsing WSDL:")) == 0) &&
                (stripos($message, "failed to load external entity") !== false)
            )
            {
                $code = self::ERROR_SIMAXONLINESECURITE;
                $message.=" Verify the uuid security.";
            }
            else
            {
                $message.= 'need php extension soap';
                $code = self::ERROR_NEEDSOAPEXT;
            }

            $this->_writeTrace($message);
            throw new \Exception($message, $code);
        }

        if (!empty($this->aOptions['apiuuid'])){
            $oSOAPClient->setSoapHeaderAPIUUID($this->aOptions['apiuuid']);
        }
        $oSOAPClient->setSoapHeaderOptionDialogue($this->oOptionDialogue);

        if (!empty($this->aOptions['token'])) {
            $stGetTokenSessionParam = new GetTokenSession($this->aOptions['token']);
        }
        else {
            $stGetTokenSessionParam = new GetTokenSession(null, $this->clUsernameToken);
        }

        $result = $oSOAPClient->GetTokenSession($stGetTokenSessionParam);
        $returnType = $oSOAPClient->getSoapHeaderReturnType();

        $this->_writeLog($oSOAPClient);
        if ($returnType != ReturnType::VALUE_IDENTIFICATION){
            $this->_writeLog($oSOAPClient);
            if (!empty($this->aOptions['token'])) {
                $message = "Token invalide";
            }
            else {
                $message = "Bad credentials ({$this->clUsernameToken->getUsername()})";
            }
            throw new \Exception($message, self::ERROR_BADCREDENTIALS);
        }

        $this->sTokenSession = $result->getSessionToken();

        if (!empty($this->aOptions['token'])) {
            $clConnectedUser = $oSOAPClient->getSoapHeaderConnectedUser();

            $clPwdInfo = $clConnectedUser->getPwdInfo();
            $this->clUsernameToken = new PartialOASISUsernameToken(
                $clConnectedUser->getElement()->getTitle(),
                $clPwdInfo,
                $this->aOptions['secret']
            );

        }

        //on a le token de session qui va nous servir par la suite
        $oSOAPClient->setSoapHeaderSessionToken($this->sTokenSession);
        $oSOAPClient->setSoapHeaderUsernameToken($this->clUsernameToken);
        $this->clSoapService =$oSOAPClient;
        return true;
    }

    /**
     * @return true
     */
    public function bDisconnect() : bool
    {
        if (!is_null($this->clSoapService)){

            if (!is_null($this->clInfoLog)){
                $this->_SendLog();
            }

            try
            {
                $this->clUsernameToken->Compute();
                $this->clSoapService->setSoapHeaderActionContext(null);
                $this->clSoapService->setSoapHeaderAutoValidate(null);
                $this->clSoapService->Disconnect(new EmptyType());
                $this->_writeLog($this->clSoapService);
            }
            catch (\Exception $e)
            {
                //on ne fait rien si ça plante sur la déconnexion
            }
            $this->clSoapService = null;
        }
        return true;
    }

    /**
     * @param       $idAction
     * @param array $aTabParams
     *
     * @return mixed|string|null
     * @throws \Exception
     */
    abstract protected function _execute($idAction, array $aTabParams);

    /**
     * @param       $idAction
     * @param array $aTabParams
     *
     * @return mixed|string|null
     * @throws \Exception
     */
    final public function execute($idAction, array $aTabParams = array())
    {
        if (!is_null($this->clInfoLog)){
            $this->clInfoLog->setInfoAction($idAction, $aTabParams);
        }

        $ret = $this->_execute($idAction, $aTabParams);
        if (!is_null($this->clInfoLog)) {
            $this->clInfoLog->return = $ret;
        }
        return $ret;
    }

    /**
     * Gère la création, la mise à jour ou la suppression d'une donnée.
     *
     * @param string $action L'action à effectuer ('create', 'update', 'delete')
     * @param string|null $idtableau L'ID de l'élément à mettre à jour ou supprimer (nécessaire pour update et delete)
     * @param array|null $data Les données à utiliser pour créer ou mettre à jour (nécessaire pour create et update)
     * @return mixed Le résultat de l'opération SOAP
     * @throws Exception Si l'action n'est pas supportée ou en cas d'erreur SOAP
     */
    public function manageData(string $action, ?string $idtableau = null, ?string $idenreg = null, ?array $data = null)
    {
        //Verification des paramètres
        if (!in_array($action, ['create', 'update', 'delete'])) {
            throw new Exception("Action non supportée : $action");
        }

        if (empty($idtableau)){
            throw new Exception("L'ID Tableau est requis pour les actions.");
        }

        if (in_array($action, ['update', 'delete']) && empty($idenreg) ) {
            throw new Exception("L'ID Enreg est requis pour les actions 'update' et 'delete'.");
        }

        if (in_array($action, ['create', 'update']) && empty($data)) {
            throw new Exception("Les données sont requises pour les actions 'create' et 'update'.");
        }

        try {
            switch ($action) {
                case 'create':
                    return $this->_executeCreate($idtableau, $data);

                case 'update':
                    return $this->_executeUpdate($idtableau, $idenreg, $data);

                case 'delete':
                    return $this->_executeDelete($idtableau, $idenreg);

                default:
                    throw new Exception("Action inconnue : $action");
            }
        } catch (Exception $e) {
            throw new Exception("Erreur lors de l'exécution de l'action $action : " . $e->getMessage());
        }
    }

    private function _executeCreate($idtableau, array $data)
    {
        // Transformation des paramètres $data en XML ou en structure SOAP acceptée
        $updateData = $this->_TabData2UpdateData($idtableau, $data);
        $stParam = new Create($idtableau, null, null, null, null, $updateData);

        //on demande la validation de la creation
        $this->clSoapService->setSoapHeaderAutoValidate(1);

        // Exécution de la requête SOAP
        $resultRequest = $this->clSoapService->Create($stParam);

        // Écriture dans le log
        $this->_writeLog($this->clSoapService);

        // Gestion du retour
        $returnType = $this->clSoapService->getSoapHeaderReturnType();
        return $this->_TraiteRetour($resultRequest, $returnType, "Création d'un élément");
    }

    private function _executeUpdate(string $idtableau, string $idenreg, array $data)
    {
        // Ajout de l'ID de l'élément à mettre à jour et transformation des données
        $paramXML = $this->_TabParams2ParamXML([$idtableau => $idenreg]);
        $updateData = $this->_TabData2UpdateData($idtableau, $data);

        $stParam = new Update($idtableau, $paramXML, null, $updateData);

        //on demande la validation de la creation
        $this->clSoapService->setSoapHeaderAutoValidate(1);

        // Exécution de la requête SOAP
        $resultRequest = $this->clSoapService->Update($stParam);

        // Écriture dans le log
        $this->_writeLog($this->clSoapService);

        // Gestion du retour
        $returnType = $this->clSoapService->getSoapHeaderReturnType();
        return $this->_TraiteRetour($resultRequest, $returnType, "Mise à jour de l'élément $idtableau");
    }

    private function _executeDelete(string $idtableau, string $idenreg)
    {
        // Ajout de l'ID de l'élément à supprimer
        $paramXML = $this->_TabParams2ParamXML([$idtableau => $idenreg]);

        $stParam = new Delete($idtableau, $paramXML);

        //on demande la validation de la creation
        $this->clSoapService->setSoapHeaderAutoValidate(1);

        // Exécution de la requête SOAP
        $resultRequest = $this->clSoapService->Delete($stParam);

        // Écriture dans le log
        $this->_writeLog($this->clSoapService);

        // Gestion du retour
        $returnType = $this->clSoapService->getSoapHeaderReturnType();
        return $this->_TraiteRetour($resultRequest, $returnType, "Suppression de l'élément $idtableau");
    }

    /**
     * @param $rep
     * @return mixed
     * @throws \Exception
     */
    protected function _ConfirmMessageBox($rep)
    {
        //contexte d'action
        $idcontext = $this->clSoapService->getSoapHeaderActionContext();
        $this->clSoapService->setSoapHeaderActionContext($idcontext);

        $stParam = new ConfirmResponse($rep);
        $resultRequest = $this->clSoapService->ConfirmResponse($stParam);
        $this->clSoapService->setSoapHeaderActionContext(null);
        $this->_writeLog($this->clSoapService);

        $returnType = $this->clSoapService->getSoapHeaderReturnType();
        return $this->_TraiteRetour($resultRequest, $returnType, "Confirmation de la message box");
    }

    /**
     * @param Element $element
     *
     * @return mixed|string|null
     * @throws \Exception
     */
    protected function _Save(Element $element)
    {
        //contexte d'actions
        $idcontext = $this->clSoapService->getSoapHeaderActionContext();
        $this->clSoapService->setSoapHeaderActionContext($idcontext);

        $resultRequest = $this->clSoapService->Validate(new Validate());
        $this->_writeLog($this->clSoapService);

        $returnType = $this->clSoapService->getSoapHeaderReturnType();
        return $this->_TraiteRetour($resultRequest, $returnType, "Enregistrement de la fiche ".$element->getTitle());
    }


    /**
     * @param Action|string $clAction
     * @param array  $aTabParams
     * @param $idcontext
     *
     * @return mixed|string|null
     * @throws \Exception
     */
    protected function _executeAction($clActionOrIDAction, array $aTabParams, $idcontext)
    {
        $idaction = $clActionOrIDAction instanceof Action ? $clActionOrIDAction->get_() : $clActionOrIDAction;

        $this->clSoapService->setSoapHeaderActionContext($idcontext);

        $stParam = new Execute($idaction, null, $this->_TabParams2ParamXML($aTabParams));
        $resultRequest = $this->clSoapService->Execute($stParam);
        //écriture dans le log
        $this->_writeLog($this->clSoapService);

        $returnType = $this->clSoapService->getSoapHeaderReturnType();

        $sOrigine = $clActionOrIDAction instanceof Action
            ? "Validation de l'action ".$clActionOrIDAction->getTitle()
            : '';

        return $this->_TraiteRetour($resultRequest, $returnType, $sOrigine);
    }

    /**
     * @param XmlType $resultRequest
     * @param string $returnType
     * @param string $sOrigine
     * @return mixed
     * @throws \Exception
     */
    protected function _TraiteRetour(XmlType $resultRequest, string $returnType, string $sOrigine='')
    {
        switch ($returnType)
        {
            case ReturnType::VALUE_EMPTY:
            {
                return null;
            }

            case ReturnType::VALUE_MESSAGE_BOX:
            {
                $action = $this->clSoapService->getSoapHeaderAction();
                $xml = simplexml_load_string($resultRequest->getXml());
                $btns = $xml->MessageBox->ButtonList->children();
                if(count($btns)>1){
                    $aTabMessage = [
                        "Une erreur est survenue lors de l'action '{$action->getTitle()}'.",
                        "La question suivante a été posée :",
                        $xml->MessageBox->Message,
                    ];
                    $this->_throwException($aTabMessage, self::ERROR_MESSAGEBOXNOTOK, $sOrigine, null);
                    break;
                }

                $nd = $btns[0];
                return $this->_ConfirmMessageBox((string)$nd);
            }

            case ReturnType::VALUE_REPORT:
            {
                $xml = simplexml_load_string($resultRequest->getXml());
                return (string)$xml->Report;
            }

            case ReturnType::VALUE_RECORD:
            {
                $element = $this->clSoapService->getSoapHeaderElement();
                return $this->_Save($element);
            }

            case ReturnType::VALUE_VALIDATE_ENREG:
            case ReturnType::VALUE_VALIDATE_RECORD:
            {
                $validateError = $this->clSoapService->getSoapHeaderValidateError();
                $action = $this->clSoapService->getSoapHeaderAction();
                $element = $this->clSoapService->getSoapHeaderElement();
                $aTabMessage = [
                    "Une erreur est survenue lors de l'action '{$action->getTitle()}', enregistrement '{$element->getTitle()}'.",
                    $validateError->getMessage(),
                ];
                $this->_throwException($aTabMessage, self::ERROR_OUVERTUREFICHE, $sOrigine, null);
                break;
            }

            case 'ERROR' :
            {
                $this->_TraiteRetour_Error($sOrigine);
                return null;
            }

            case ReturnType::VALUE_VALIDATE_ACTION:
            {
                return $this->_TraiteRetour_ValidateAction($resultRequest, $sOrigine);
            }
            case ReturnType::VALUE_AMBIGUOUS_ACTION:
            default:
            {
                $action = $this->clSoapService->getSoapHeaderAction();
                $aTabMessage = [
                    "Une erreur est survenue lors de l'action '{$action->getTitle()}'",
                    "Le type retour $returnType n'est pas géré.",
                ];
                $this->_throwException($aTabMessage, self::ERROR_TRAITERETOUR, $sOrigine, null);
            }
        }
    }

    /**
     * @param string $sOrigine
     * @return void
     * @throws \Exception
     */
    private function _TraiteRetour_Error(string $sOrigine)
    {
        $clEnvelope = simplexml_load_string($this->clSoapService->getLastResponse());
        //calcul du nom du namespace de l'enveloppe
        $tabNamespaceDuXML    = $clEnvelope->getNamespaces();
        // tableau associatif nom => url

        // Il n'y a que ces 4 là qui soit possibles pour l'enveloppe SOAP (standard w3c)
        // les deux même avec / final ou nom pour compatibilité avec vieille version de NOUTOnline
        $aNamespacePossible = array(
            'http://www.w3.org/2003/05/soap-envelope/',
            'http://schemas.xmlsoap.org/soap/envelope/',
            'http://www.w3.org/2003/05/soap-envelope',
            'http://schemas.xmlsoap.org/soap/envelope',
        );

        //recherche les namespaces possible dans les namespaces du XML
        foreach($aNamespacePossible as $sUrlNamespacePossible)
        {
            $sNamespaceSOAP = array_search($sUrlNamespacePossible, $tabNamespaceDuXML); // false si ne trouve pas la valeur dans le tableau
            //renvoi la clé si trouvé, donc le nom du namespace

            if ($sNamespaceSOAP !== false) // Si le namespace de l'enveloppe est un de ceux autorisés
            {
                break;
            }
        }

        $ndBody   = $clEnvelope->children($sNamespaceSOAP, true)->Body;

        /* @var $ndFault \SimpleXMLElement */
        $ndFault = $ndBody->children($sNamespaceSOAP, true)->Fault;
//le noeud ListErr fils de  Detail
        /* @var $ndListErr \SimpleXMLElement */
        $ndDetail = $ndFault->children($sNamespaceSOAP, true)->Detail;
        if (!empty($ndDetail->children())){
            $ndListErr = $ndDetail->children()->ListErr;
        }
        else {
            $ndListErr = simplexml_load_string((string)$ndDetail);
        }


        //on recherche le namespace pour les erreurs SIMAX http://www.nout.fr/soap/error
        $messages = [];
        $ndErrors = $ndListErr->children('http://www.nout.fr/soap/error');
        foreach ($ndErrors as $ndError)
        {
            /* @var $ndError \SimpleXMLElement */
            $messages[] = (string)$ndError->children()->Message;
        }

        $this->_throwException($messages, self::ERROR_TRAITERETOUR, $sOrigine, null);
    }

    /**
     * @param $resultRequest
     * @param string $sOrigine
     *
     * @return mixed|string|null
     * @throws \Exception
     */
    private function _TraiteRetour_ValidateAction($resultRequest, string $sOrigine)
    {

        //on regarde s'il y a des paramètres obligatoire non remplie
        $clAction = $this->clSoapService->getSoapHeaderAction();
        $xsdSchema = $this->clSoapService->getSoapHeaderXSDSchema();
        $ndSchema = simplexml_load_string($xsdSchema);
        $ndXML = simplexml_load_string($resultRequest->getXml());
        $required = $ndSchema->xpath('//*[@xs:use="required"]');

        $emptyRequired = [];
        foreach ($required as $node)
        {
            $balise = (string)($node->attributes('http://www.w3.org/2001/XMLSchema')['name']);
            $libelle = (string)($node->attributes('http://www.nout.fr/XMLSchema')['name']);

            $xpathret = $ndXML->xpath('//'.$balise);
            $ndRequired = $xpathret[0];

            $aSIMAXAttribs = $ndRequired->attributes('simax', true);
            if ($aSIMAXAttribs['ref']){
                $val = $this->_sGetValBaliseData($aSIMAXAttribs['ref'], $ndXML);
            }
            else {
                $val = (string)$ndRequired;
            }

            if (empty($val)){
                $emptyRequired[]="- ".$libelle;
            }
        }

        if (!empty($emptyRequired)){
            $message = "Une erreur est survenue lors de l'action '{$clAction->getTitle()}', \r\nles paramètres obligatoires suivant ne sont pas remplis :\r\n"
                       .implode("\r\n", $emptyRequired);
            $this->_throwException($message, self::ERROR_OUVERTUREPARAMETRE, $sOrigine, null);
        }

        //il n'y a pas de paramètre obligatoire non rempli

        $form = $this->clSoapService->getSoapHeaderForm();
        $idForm = $form->get_();

        $idcontext = $this->clSoapService->getSoapHeaderActionContext();
        $aTabParams = [];
        foreach( $ndXML->{'id_'.$idForm}->children() as $ndParam)
        {
            $name = $ndParam->getName();
            preg_match_all('/id_(?<id>\d+)/', $name, $matches, PREG_SET_ORDER, 0);

            $aSIMAXAttribs = $ndParam->attributes('simax', true);
            if ($aSIMAXAttribs['ref']){
                //balise Data, il faut trouver le noeud qui va bien
                $aTabParams[$matches[0]['id']]=$this->_sGetValBaliseData($aSIMAXAttribs['ref'], $ndXML);
            }
            else {
                $aTabParams[$matches[0]['id']]=(string)$ndParam;
            }
        }

        return $this->_executeAction($clAction, $aTabParams,$idcontext);
    }

    protected function _sGetValBaliseData($ref, $ndXML) : string
    {
        $ndXML->registerXPathNamespace('d', 'http://www.nout.fr/XML/');
        $xpathdata = $ndXML->xpath('//d:Data[@d:ref="'.$ref.'"]');
        $ndData = $xpathdata[0];
        $val = (string)$ndData;
        $encoding = $ndData->attributes('simax', true)["encoding"] ?? '';
        if ($encoding == "quoted-printable"){
            $val = mb_convert_encoding(
                quoted_printable_decode($val)
                , 'UTF-8'
                , 'ISO-8859-1'
            ) ;
        }
        return $val;
    }


    /**
     * @param            $message
     * @param            $code
     * @param \Exception $e
     * @param string $sOrigine
     *
     * @throws \Exception
     */
    protected function _throwException($message, $code, string $sOrigine, ?\Exception $e)
    {
        if (is_array($message)){
            $message = implode("\r\n", $message);
        }
        if (!empty($sOrigine)){
            $message = "Erreur sur : $sOrigine.\r\n$message";
        }
        if (!empty($this->clInfoLog))
        {
            $lastRequest = end($this->clInfoLog->requestStack);
            $lastRequest->exception = new \stdClass();
            $lastRequest->exception->code = $code;
            $lastRequest->exception->message = $message;
        }
        throw new \Exception($message, $code, $e);
    }

    /**
     * @return void
     */
    protected function _SendLog()
    {
        $stParam = new Execute($this->clActionLog->actionID, null, $this->_TabParams2ParamXML([
            $this->clActionLog->parameterID => json_encode($this->clInfoLog)
        ]));
        try {
            //bouteille à la mer
            $this->clUsernameToken->Compute();
            $this->clSoapService->setSoapHeaderActionContext(null);
            $this->clSoapService->setSoapHeaderAutoValidate(null);
            $this->clSoapService->Execute($stParam);
        }
        catch (\Exception $e)
        {
            //on ne fait rien sur le catch, pas important à cet endroit
        }
    }


    /**
     * @param DebugTrace $obj
     */
    public function setDebugTrace(DebugTrace $obj)
    {
        $this->sRequestID = $obj->getUniqueId();
        $this->clDebugTrace = $obj;
        $this->clDebugTrace->setHostAndPort($this->aOptions['host'], $this->aOptions['port']);
    }

    /**
     * @param $idaction
     * @param $idparameter
     */
    public function setLogSIMAX($idaction, $idparameter)
    {
        $this->clActionLog = new ActionLog($idaction, $idparameter);
        $this->clInfoLog   = new InfoLog($this->sRequestID);
    }


    /**
     * @param $trace
     */
    protected function _writeTrace($trace)
    {
        if (!$this->clDebugTrace)
        {
            return;
        }
        $this->clDebugTrace->writeTrace($trace);
    }

    /**
     * @param SIMAXService $service
     */
    protected function _writeLog(SIMAXService $service)
    {
        if ($this->clDebugTrace || !is_null($this->clInfoLog)){
            $requestHeaders = $service->getLastRequestHeaders();
            $responseHeaders = $service->getLastResponseHeaders();

            $requestBody=$service->getLastRequest();
            $responseBody=$service->getLastResponse();

            $request = <<<EOT
$requestHeaders
$requestBody
EOT;
            $response = <<<EOT
$responseHeaders
$responseBody
EOT;

            if ($this->clDebugTrace){
                $this->clDebugTrace->writeTrace('request : ');
                $this->clDebugTrace->writeTrace($request);
                $this->clDebugTrace->writeTrace("~~~~~~~~~~~~~~~");
                $this->clDebugTrace->writeTrace('response : ');
                $this->clDebugTrace->writeTrace($response);
            }

            if(!is_null($this->clInfoLog)){
                $this->clInfoLog->requestStack[] =new RequestLog($requestBody, $responseBody);
            }
        }
    }

    /**
     * @param array $TabParams
     * @return string
     */
    protected function _TabParams2ParamXML(array $TabParams) : string
    {

        $paramXML = '';
        foreach($TabParams as $idcolonne=>$valeur)
        {
            $idcolonne = strpos($idcolonne, 'id_') !== false ? $idcolonne : "id_$idcolonne";

            $paramXML.="<$idcolonne>".htmlspecialchars($valeur, ENT_XML1 | ENT_QUOTES | ENT_SUBSTITUTE)."</$idcolonne>";
        }
        return $paramXML;
    }


    /**
     * @param array $TabParams
     * @return string
     */
    protected function _TabData2UpdateData(string $idtableau, array $TabParams) : string
    {
        $idtableau = strpos($idtableau, 'id_') !== false ? $idtableau : "id_$idtableau";

        $updateData = "<xml><$idtableau>";
        foreach($TabParams as $idcolonne=>$valeur)
        {
            $idcolonne = strpos($idcolonne, 'id_') !== false ? $idcolonne : "id_$idcolonne";

            $updateData.="<$idcolonne>".htmlspecialchars($valeur, ENT_XML1 | ENT_QUOTES | ENT_SUBSTITUTE)."</$idcolonne>";
        }
        $updateData.="</$idtableau></xml>";
        return $updateData;
    }



    /**
     * @param \SimpleXMLElement $clMessageBoxNode
     * @return false|mixed
     */
    protected function _GetJSONFromMessageBox(\SimpleXMLElement $clMessageBoxNode)
    {
        if (    (count($clMessageBoxNode->ButtonList->Children()->TypeConfirmation)!=1)
            ||  ((int)$clMessageBoxNode->ButtonList->TypeConfirmation!=ConfirmResponse::IDOK)){
            return [false, false, null, null];
        }

        $message = trim((string)$clMessageBoxNode->Message);
        if (($message[0] != '{') && ($message[0]!='[')){
            return [false, false, null, null];
        }

        $error = null;
        try {
            $json = json_decode($message, null, 512, JSON_THROW_ON_ERROR|JSON_BIGINT_AS_STRING);
        }
        catch (\Exception $e)
        {
            $error = $e->getMessage();
        }

        return [true, $json, $error, $message];
    }

}
